-------------------------------------------------------
-- Design Name : in_del_1x_03
-- File Name   : in_del_1x_03.vhd
-- Device      : Spartan 6
-- Function    : setting the input delay for 1 channel
-- Coder       : K.-H. Sulanke, DESY, 2018-03-19
-------------------------------------------------------
-- 2017-02-14  del_done replaced by del_busy
-- 2017-02-24  stm restart from any state, if new del_load ocurred
--             note, the time between consecutive del_load's is given by the SPI write cycle time
-- 2018-03-19  ser. data input as single signal now

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all ;

library unisim ;
use unisim.vcomponents.all ;

entity in_del_1x_03 is 
  port 	(
        reset			            :  in std_logic ;   -- reset (active high)
        clk                   :  in std_logic ;   -- global clock input
        io_clk                :  in std_logic ;   -- high speed i/o clock
        sdat_in  	            :  in std_logic;   -- lvds serial data inputs
        del_data              :  in std_logic_vector (7 downto 0);  -- delay value
        del_load              :  in std_logic;   -- to initiate the DELAY cycle, single pulse
        del_busy              :  out std_logic;  -- delay done, handshake with del_load       
        del_dataout2          :  out std_logic   -- 
        );
  end in_del_1x_03 ;

architecture arch_in_del_1x_03 of in_del_1x_03 is

  signal cal          : std_logic :='0';
  signal rst            : std_logic :='0';
  signal iodelay_busy   : std_logic :='0';
  --signal busy_0         : std_logic :='0';  -- needed for synchronization pipeline
  signal busy           : std_logic :='0';
  signal ce             : std_logic :='0';
  signal inc            : std_logic :='0';
  signal del_count      : std_logic_vector(7 downto 0);
  signal del_load0      : std_logic :='0';  -- by one clock delayed del_load signal
  
  type   state_type is (DEL_IDLE, DEL_CAL_WAIT, DEL_CAL, DEL_RST_WAIT, DEL_RST, DEL_CE_OFF, DEL_INC, DEL_RDY, DEL_BREAK, DEL_RESTART);
  signal state : state_type := DEL_IDLE;
     
 
begin


-------------------------------- INPUT_DELAY_inst instantiation -----------------------------------------
 


iodelay_m : IODELAY2 generic map(
	DATA_RATE      		=> "SDR", 			-- <SDR>, DDR
	SIM_TAPDELAY_VALUE	=> 50,  			-- nominal tap delay (sim parameter only)
	IDELAY_VALUE  		=> 0, 				-- {0 ... 255}
	IDELAY2_VALUE 		=> 0, 				-- {0 ... 255}
	ODELAY_VALUE  		=> 0, 				-- {0 ... 255}
	IDELAY_MODE   		=> "NORMAL", 			-- "NORMAL", "PCI"
	SERDES_MODE   		=> "NONE", 			-- <NONE>, MASTER, SLAVE
	IDELAY_TYPE   		=> "VARIABLE_FROM_ZERO", 	-- "DEFAULT", "DIFF_PHASE_DETECTOR", "FIXED", "VARIABLE_FROM_HALF_MAX", "VARIABLE_FROM_ZERO"
	COUNTER_WRAPAROUND 	=> "STAY_AT_LIMIT", 		-- STAY_AT_LIMIT, WRAPAROUND
	DELAY_SRC     		=> "IDATAIN" )			-- "IO", "IDATAIN", "ODATAIN"
port map (
	IDATAIN  		=> sdat_in, 		     -- input, data from master IOB
	TOUT     		=> open, 			  -- output, tri-state signal to IOB
	DOUT     		=> open, 			  -- output, data to IOB
	T        		=> '1', 			     -- input, tri-state control from OLOGIC/OSERDES2, tie high fro input only mode
	ODATAIN  		=> '0', 			     -- input, data from OLOGIC/OSERDES2
	DATAOUT  		=> open,--sdat_dld_m, 	  -- output, delayed data 1 to ILOGIC/ISERDES2
	DATAOUT2 		=> del_dataout2,--open, 			  -- output, data 2 to ILOGIC/ISERDES2, for PCI ?
	IOCLK0   		=> io_clk, 			  -- input, High speed clock for calibration
	IOCLK1   		=> '0', 			     -- input, High speed clock for calibration
	CLK      		=> clk,         -- input, Fabric clock (GCLK) for control signals
	CAL      		=> cal, 	  -- input, Calibrate enable signal
	INC      		=> inc, 			     -- input, increment / decrement delay counter
	CE       		=> ce, 			     -- input, enable increment / decrement delay counter
	RST      		=> rst,       -- input, Reset delay line to 1/2 max in this case
	BUSY        => iodelay_busy) ;         -- output signal indicating sync circuit has finished / calibration has finished



-- controlling the individual input delays ----------------------------------------------- 

 SET_INPUT_DELAY: process (clk)
    begin
     if clk'event and clk = '1' then
       if reset = '1' then
         state <= DEL_IDLE;
         cal       <= '0';
         inc       <= '0';
         ce        <= '0';
         rst       <= '0';
         del_busy  <= '0';
         del_load0 <= '0';
         busy      <= '0';
       else
         del_load0 <= del_load;
         busy      <= iodelay_busy;
         
   DEL_SM : case (state) is
   
          when DEL_IDLE => 
            cal       <= '0';
            inc       <= '0';
            ce        <= '0';
            rst       <= '0';
            del_busy  <= '0';
            if del_load0 = '1' then          -- delayed del_load signal
              del_busy <= '1';
              cal      <= '1';
              state    <= DEL_CAL_WAIT;
            end if;

          when DEL_BREAK =>                -- break by a new del_load
            cal       <= '0';
            inc       <= '0';
            ce        <= '0';
            rst       <= '0';
            del_busy  <= '0';
            state     <= DEL_RESTART;

          when DEL_RESTART => 
             del_busy <= '1';
             cal      <= '1';
             state    <= DEL_CAL_WAIT;
             
          when DEL_CAL_WAIT =>                 -- wait for the command being accepted
            if del_load0 = '1' then
              state <= DEL_BREAK;
            elsif busy = '1' then 
              cal   <= '0';
              state <= DEL_CAL;
            end if;  
              
          when DEL_CAL =>                      -- wait for calbration-readiness
            if del_load0 = '1' then
              state <= DEL_BREAK;
            elsif busy = '0' then
              rst   <= '1';             
              state <= DEL_RST_WAIT;
            end if; 

         when DEL_RST_WAIT =>                 -- wait for the command being accepted
            if del_load0 = '1' then
              state <= DEL_BREAK;
            elsif busy = '1' then 
              state <= DEL_RST;
            end if;  
            
          when DEL_RST =>                       -- reset the delay channel
            rst <= '0';  
            if del_load0 = '1' then
              state <= DEL_BREAK;
            elsif busy = '0' then
              del_count <= del_data;
              ce    <= '1';
              inc   <= '1';
              state <= DEL_CE_OFF;
            end if;    

          when DEL_CE_OFF =>
              ce    <= '0';
              state <= DEL_INC;

          when DEL_INC =>
            if del_load0 = '1' then
              state <= DEL_BREAK;
            elsif busy = '0' then
              if del_count = X"00" then
                del_busy   <= '0';
                inc        <= '0';
                ce         <= '0';               
                state      <= DEL_RDY;
              else
               del_count <= del_count - '1';
               ce    <= '1';
               state <= DEL_CE_OFF;
              end if;
            end if;    

          when DEL_RDY => 
            del_busy  <= '0';
            if del_load0 = '1' then          -- delayed del_load signal
              del_busy <= '1';
              inc      <= '1';              
              cal      <= '1';
              state    <= DEL_CAL_WAIT;
            end if;  

        end case DEL_SM;
      end if;
    end if;
   end process SET_INPUT_DELAY; 

end arch_in_del_1x_03;