-------------------------------------------------------
-- Design Name : in_del_w_cal_01
-- File Name   : in_del_w_cal_01.vhd
-- Device      : Spartan 6
-- Function    : adjusting the input delay for a channel (TIMEPPS, L1A)
--               either by software or by calibration
-- Coder       : K.-H. Sulanke, DESY, 2017-10-16
-------------------------------------------------------
-- 2017-03-23  first version
-- 2017-10-16  cal_cyc_nd, del_cyc_nd  commented

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all ;

library unisim ;
use unisim.vcomponents.all ;

entity in_del_w_cal_01 is 
  port 	(
        reset			            :  in std_logic ;   -- reset (active high)
        clk                   :  in std_logic ;   -- global clock input
        iodelay_ioclk0        :  in std_logic ;   -- high speed i/o delay clock
        sdat_in_p	            :  in std_logic;   -- lvds serial data inputs
        sdat_in_n             :  in std_logic;
        del_data              :  in std_logic_vector (7 downto 0);  -- delay value
        del_load              :  in std_logic;   -- to initiate the DELAY cycle, single pulse
        del_busy              :  out std_logic;  -- delay done, handshake with del_load
        cal_start             :  in std_logic;   -- single pulse to initiate the calibration cycle
        cal_iodelay_rst       :  in std_logic;   -- iodelay_rst, controlled by phase detecting unit
        cal_iodelay_cal       :  in std_logic;   -- iodelay_cal, controlled by phase detecting unit
        cal_iodelay_ce        :  in std_logic;   -- iodelay_ce, controlled by phase detecting unit
        cal_iodelay_inc       :  in std_logic;   -- iodelay_inc, controlled by phase detecting unit
        cal_iodelay_busy      :  out std_logic;  -- iodelay_busy, needed by phase detecting unit
        cal_data              :  out std_logic_vector (7 downto 0);  -- gained delay value, after calibration
        cal_err               :  in std_logic;   -- single pulse
        cal_done              :  in std_logic;   -- single pulse
        --cal_cyc_nd            :  out std_logic;   --
        --del_cyc_nd            :  out std_logic;   --
        iodelay_dataout2      :  out std_logic    -- delayed sdat_in signal
        );
  end in_del_w_cal_01 ;

architecture arch_in_del_w_cal_01 of in_del_w_cal_01 is
  
  constant WAIT_TIME    : std_logic_vector := B"101";  -- amount of clk's, should be greater 4
  signal busy_wait_ct   : std_logic_vector(WAIT_TIME'length -1 downto 0); -- extra wait time after iodelay_busy='0'
  
	signal sdat_in        : std_logic;  -- input, 
  signal iodelay_cal    : std_logic :='0';
  signal iodelay_rst    : std_logic :='0';
  signal iodelay_ce     : std_logic :='0';
  signal iodelay_inc    : std_logic :='0';
  signal iodelay_busy   : std_logic :='0';
  signal rst            : std_logic :='0';
  signal busy           : std_logic :='0';
  signal cal            : std_logic :='0';
  signal ce             : std_logic :='0';
  signal inc            : std_logic :='0';
  signal del_count      : std_logic_vector(7 downto 0);
  signal del_adj        : std_logic :='0';  -- by one clock delayed del_load signal
  
  signal cal_cyc        : std_logic :='0';  -- calibration cycle is running 
  signal del_cyc        : std_logic :='0';  -- delay set cycle is running 
  signal del_done       : std_logic :='0';  -- delay set cycle done, single pulse
  
  type   state_type is (DEL_IDLE, DEL_CAL_WAIT, DEL_CAL, DEL_RST_WAIT, DEL_RST, CAL_CYCLE, DEL_CE_OFF, DEL_INC, DEL_BREAK, DEL_RESTART);
  signal state : state_type := DEL_IDLE;
     
 
begin


-------------------------------- INPUT_DELAY_inst instantiation -----------------------------------------
 

iob_data_in : IBUFDS generic map(
	DIFF_TERM		=> true)
port map (
	I    			=> sdat_in_p,
	IB       	=> sdat_in_n,
	O         => sdat_in);


iodelay_m : IODELAY2 generic map(
	DATA_RATE      		=> "SDR", 			-- <SDR>, DDR
	SIM_TAPDELAY_VALUE	=> 37,  			-- nominal tap delay (sim parameter only)
	IDELAY_VALUE  		=> 0, 				-- {0 ... 255}
	IDELAY2_VALUE 		=> 0, 				-- {0 ... 255}
	ODELAY_VALUE  		=> 0, 				-- {0 ... 255}
	IDELAY_MODE   		=> "NORMAL", 			-- "NORMAL", "PCI"
	SERDES_MODE   		=> "NONE", 			-- <NONE>, MASTER, SLAVE
	IDELAY_TYPE   		=> "VARIABLE_FROM_ZERO", 	-- "DEFAULT", "DIFF_PHASE_DETECTOR", "FIXED", "VARIABLE_FROM_HALF_MAX", "VARIABLE_FROM_ZERO"
	COUNTER_WRAPAROUND 	=> "STAY_AT_LIMIT", 		-- STAY_AT_LIMIT, WRAPAROUND
	DELAY_SRC     		=> "IDATAIN" )			-- "IO", "IDATAIN", "ODATAIN"
port map (
	IDATAIN  		=> sdat_in, 		     -- input, data from master IOB
	TOUT     		=> open, 			  -- output, tri-state signal to IOB
	DOUT     		=> open, 			  -- output, data to IOB
	T        		=> '1', 			     -- input, tri-state control from OLOGIC/OSERDES2, tie high fro input only mode
	ODATAIN  		=> '0', 			     -- input, data from OLOGIC/OSERDES2
	DATAOUT  		=> open,--sdat_dld_m, 	  -- output, delayed data 1 to ILOGIC/ISERDES2
	DATAOUT2 		=> iodelay_dataout2,--open, 			  -- output, data 2 to ILOGIC/ISERDES2, for PCI ?
	IOCLK0   		=> iodelay_ioclk0, 			  -- input, High speed clock for calibration
	IOCLK1   		=> '0', 			     -- input, High speed clock for calibration
	CLK      		=> clk,         -- input, Fabric clock (GCLK) for control signals
	CAL      		=> iodelay_cal, 	  -- input, Calibrate enable signal
	INC      		=> iodelay_inc, 			     -- input, increment / decrement delay counter
	CE       		=> iodelay_ce, 			     -- input, enable increment / decrement delay counter
	RST      		=> iodelay_rst,       -- input, Reset delay line to 1/2 max in this case
	BUSY        => iodelay_busy) ;         -- output signal indicating sync circuit has finished / calibration has finished



-- controlling the individual input delays ----------------------------------------------- 
 
 
 DEL_CYCLE_ARBITER : process (clk)
   begin
     if clk'event and clk = '1' then
       if reset = '1' then
        del_adj  <= '0';
        del_cyc  <= '0';
        cal_cyc  <= '0';
        del_busy <= '0';    -- the del_load or cal_start pulse (re)starts the delay set cycle always
       else
        if del_load = '1' then  -- by software, has higher priority
         del_busy <= '1';      -- software status bit
         del_cyc  <= '1';      -- extend the single pulse
         cal_cyc  <= '0';      -- stop never ending cal cycle 
        elsif del_cyc = '1' and del_done = '1' then
         del_busy <= '0';
         del_cyc  <= '0';
        end if;
        if cal_start = '1' and del_load = '0' then  -- by hand, has lower priority
         del_busy <= '1';      -- software status bit
         del_cyc  <= '0';      -- stop never ending del cycle  
         cal_cyc  <= '1';      -- extend the single pulse
        elsif cal_cyc = '1' and (cal_done = '1' or cal_err = '1') then
         del_busy <= '0';
         cal_cyc  <= '0'; 
        end if; -- reset = '1'
       del_adj <= '0';      -- to get single pulses
       if del_load = '1' or cal_start = '1' then
        del_adj <= '1';       -- also needed for restart from any state, see following stm
       end if;
      end if; -- reset = '1'       
     end if; -- clk'event and clk = '1'
   end process DEL_CYCLE_ARBITER;      



 SET_INPUT_DELAY: process (clk)
    begin
     if clk'event and clk = '1' then
       if reset = '1' then
         state       <= DEL_IDLE;
         del_done    <= '0';
         rst         <= '0';
         cal         <= '0';
         inc         <= '0';
         ce          <= '0';
         cal_data    <= X"00";
       else
         
   DEL_SM : case (state) is
   
          when DEL_IDLE => 
            del_done    <= '0';
            cal         <= '0';
            inc         <= '0';
            ce          <= '0';
            rst         <= '0';
            if del_adj = '1' then    -- by software, has higher priority
              cal       <= '1';
              state     <= DEL_CAL_WAIT;
            end if;

          when DEL_BREAK =>                -- break by a new del_load
            cal          <= '0';
            inc          <= '0';
            ce           <= '0';
            rst          <= '0';
            busy_wait_ct <= WAIT_TIME;
            state        <= DEL_RESTART; 

          when DEL_RESTART => 
            if busy_wait_ct = 0 then
             cal     <= '1';
             state   <= DEL_CAL_WAIT;
            else
             busy_wait_ct <= busy_wait_ct - 1;          
            end if; 
             
          when DEL_CAL_WAIT =>                 -- wait for the command being accepted
            if del_adj = '1' then
              state <= DEL_BREAK;
            elsif busy = '1' then 
              cal   <= '0';
              state <= DEL_CAL;
            end if;  
              
          when DEL_CAL =>                      -- wait for calbration-readiness
            if del_adj = '1' then
              state       <= DEL_BREAK;
            elsif busy = '0' then
              rst         <= '1';             
              state       <= DEL_RST_WAIT;
            end if; 

          when DEL_RST_WAIT =>                 -- wait for the command being accepted
            if del_adj = '1' then
              state   <= DEL_BREAK;
            elsif busy = '1' then 
              state   <= DEL_RST;
            end if;  
            
          when DEL_RST =>                       -- reset the delay channel
            rst         <= '0';  
            if del_adj = '1' then
              state   <= DEL_BREAK;
            elsif busy = '0' then
              if del_cyc = '1' then
                del_count <= del_data;
                ce        <= '1';
                inc       <= '1';
                state     <= DEL_CE_OFF;
              else
                del_count <= X"00";
                state     <= CAL_CYCLE;
              end if;
             end if;  

           when CAL_CYCLE =>                  -- ce, inc, busy handling by calibration module now
             if del_adj = '1' then
              state <= DEL_BREAK;
             elsif cal_iodelay_ce ='1' then   -- single pulse always
               if    cal_iodelay_inc = '1' and del_count /= X"ff" then
                 del_count <= del_count + 1;
               elsif cal_iodelay_inc = '0' and del_count /= X"00" then
                 del_count <= del_count - 1;
               end if;                 
             elsif cal_err = '1' then
              state    <= DEL_IDLE;
             elsif cal_done = '1' then
              cal_data <= del_count;
              state    <= DEL_IDLE;
             end if;             

           when DEL_CE_OFF =>
              ce           <= '0';
              busy_wait_ct <= WAIT_TIME; -- required for clean ce to busy handling
              state        <= DEL_INC;

           when DEL_INC =>
            if del_adj = '1' then
              state <= DEL_BREAK;
            elsif busy = '0' then
             if busy_wait_ct = 0 then      
               if del_count = X"00" then
                inc        <= '0';
                ce         <= '0';
                del_done   <= '1';                
                state      <= DEL_IDLE;
               else               
                del_count  <= del_count - '1'; 
                ce         <= '1';
                state      <= DEL_CE_OFF;
               end if; 
              else
               busy_wait_ct <= busy_wait_ct -1 ;             
              end if; --busy_wait_ct = 0
            end if; --del_adj = '1'  

        end case DEL_SM;
      end if;
    end if;
   end process SET_INPUT_DELAY;

   IODELAY_MUX:process(del_cyc, cal_cyc, ce, inc, cal, rst, iodelay_busy,
                       cal_iodelay_ce, cal_iodelay_inc, cal_iodelay_cal, cal_iodelay_rst)
    begin     
     cal_iodelay_busy  <= iodelay_busy and cal_cyc;
     busy              <= iodelay_busy;     
     iodelay_ce        <= (del_cyc and ce)  or (cal_cyc and cal_iodelay_ce);
     iodelay_inc       <= (del_cyc and inc) or (cal_cyc and cal_iodelay_inc);
     iodelay_cal       <= (del_cyc and cal) or (cal_cyc and cal_iodelay_cal);
     iodelay_rst       <= (del_cyc and rst) or (cal_cyc and cal_iodelay_rst);
     
    end process IODELAY_MUX;
  
     --cal_cyc_nd        <= cal_cyc ;  
     --del_cyc_nd        <= del_cyc ; 
  
end arch_in_del_w_cal_01;