-------------------------------------------------------
-- Design Name : iserdes_ph_ctrl_02
-- File Name   : iserdes_ph_ctrl_02.vhd
-- Device      : Spartan 6
-- Function    : iserdes two channel phase comparison and delay control for 
--               equivalent leading edge times
-- Coder       : K.-H. Sulanke, DESY, 2017-10-16
-------------------------------------------------------
-- 2017-03-28, modified in_del_serdes_pd_01.vhd
-- 2017-04-02, signal inc_plus implemented, for final adjustment
-- 2017-10-16, ref_iodelay_m/s removed


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all ;

library unisim ;
use unisim.vcomponents.all ;

entity iserdes_ph_ctrl_02 is  
 generic (
          S               : integer          := 8;	           -- to set the serdes factor 1..8
          ALIGNMENT_CNT   : std_logic_vector := B"0001_1111";  -- required max. successfull phase alignments, before calibration is done
          EXTRA_INC_CNT   : std_logic_vector := B"0000_0110"   -- for final adjustment
         );        -- clock period (ns) of input clock on clkin_p
  port (
        reset 		            :  in std_logic ;   -- reset (active high)
        clk                   :  in std_logic ;   -- global clock input
        del_ioclk0            :  in std_logic ;   -- fast (5 x clk), needed by iodelay instance
        ref_sdat_in_p         :  in std_logic;   -- lvds serial data inputs, reference signal
        ref_sdat_in_n	        :  in std_logic;   -- 
        cal_sdat_in_p         :  in std_logic;   -- lvds serial data inputs, signal with its delay to be calibrated
        cal_sdat_in_n	        :  in std_logic;   -- 
        iserdes_ioclk         :  in std_logic ;   -- high speed i/o clock
        iserdes_stb           :  in std_logic ;   -- 
        cal_iodelay_rst       :  out std_logic;   -- iodelay_rst, to control the delay of the signal to be calibrated
        cal_iodelay_cal       :  out std_logic;   -- iodelay_cal, to control the delay of the signal to be calibrated
        cal_iodelay_ce        :  out std_logic;   -- iodelay_ce, to control the delay of the signal to be calibrated
        cal_iodelay_inc       :  out std_logic;   -- iodelay_inc, to control the delay of the signal to be calibrated
        cal_iodelay_busy      :  in std_logic;    -- iodelay_busy, from the signal to be calibrated
        cal_start             :  in std_logic ;   -- start calibration cycle, single pulse
        cal_busy              :  out std_logic;   -- one while calibrating
        cal_done              :  out std_logic;   -- single pulse
        cal_err               :  out std_logic    -- calibration error, gets cleared by next cal_start
        );
  end iserdes_ph_ctrl_02 ;

architecture arch_iserdes_ph_ctrl_02 of iserdes_ph_ctrl_02 is

  signal iserdes_rst         : std_logic :='0';
 
	signal ref_sdat_in         : std_logic;
	signal ref_sdat_dld_m      : std_logic;
  signal ref_sdat_dld_s      : std_logic;
  signal ref_sdat_cascade    : std_logic :='0';  -- 
  signal ref_pd_edge         : std_logic :='0';  --     
  signal ref_iserdes_pdat    : std_logic_vector (0 to 7) := X"00";
  signal ref_iserdes_pdat_p7 : std_logic :='0';  -- previous bit 7
  signal ref_pulse_pdat      : std_logic_vector (0 to 7) := X"00";  
  signal ref_pulse_pdat_prev : std_logic_vector (0 to 3) := X"0"; 
  signal ref_pulse_pdat_incl_prev  : std_logic_vector (0 to 11);
  
	signal cal_sdat_in         : std_logic;
	signal cal_sdat_dld_m      : std_logic;
  signal cal_sdat_dld_s      : std_logic;
  signal cal_sdat_cascade    : std_logic :='0';  -- 
  signal cal_pd_edge         : std_logic :='0';  --     
  signal cal_iserdes_pdat    : std_logic_vector (0 to 7) := X"00";
  signal cal_iserdes_pdat_p7 : std_logic :='0';  -- previous bit 7
  signal cal_pulse_pdat      : std_logic_vector (0 to 7) := X"00";  
  signal cal_pulse_pdat_prev : std_logic_vector (0 to 3) := X"0"; 
  signal cal_pulse_pdat_incl_prev  : std_logic_vector (0 to 11);
  
  signal iodelay_rst        : std_logic := '0';
  signal iodelay_inc        : std_logic := '0';
  signal iodelay_ce         : std_logic := '0';
  signal iodelay_cal        : std_logic := '0';
  signal iodelay_busy       : std_logic := '0';
 
-- iserdes control
  signal del_count          : std_logic_vector(7 downto 0);
  signal busy_wait_ct       : std_logic_vector(2 downto 0);
  signal in_phase_ct        : std_logic_vector(ALIGNMENT_CNT'range);
  signal inc_plus           : std_logic := '0'; -- for final adjustment

  signal pulse_ena          : std_logic := '0';
  
  signal inc_delay          : std_logic := '0';
  signal dec_delay          : std_logic := '0';
  signal phase_aligned      : std_logic := '0';
   
  type pulse_state_type is (PULSE_IDLE, INCDEC_GEN);
   signal pulse_state : pulse_state_type := PULSE_IDLE;
   
  type state_type is (CAL_IDLE, DEL_CAL_WAIT, DEL_CAL, DEL_RST,CAL_INC_PLUS,
                      CAL_BUSY_ON, CAL_BUSY_WAIT, CAL_DEL_ADJ_WAIT, CAL_PHASE_CHECK);
   signal state : state_type := CAL_IDLE; 
 
  
begin

 
  ref_data_in : IBUFDS generic map(
    DIFF_TERM		=> true)
  port map (
    I    			=> ref_sdat_in_p,
    IB       	=> ref_sdat_in_n,
    O         => ref_sdat_in);
    
  cal_data_in : IBUFDS generic map(
    DIFF_TERM		=> true)
  port map (
    I    			=> cal_sdat_in_p,
    IB       	=> cal_sdat_in_n,
    O         => cal_sdat_in);
    
  iserdes_rst <= reset;

--- iodelay and iserdes instantiation for the reference signal -----------------------------------------------------------

--  ref_iodelay_m : IODELAY2 generic map(
--    DATA_RATE      		=> "SDR", 			-- <SDR>, DDR
--    SIM_TAPDELAY_VALUE	=> 37,  			-- nominal tap delay (sim parameter only)
--    IDELAY_VALUE  		=> 1, 				-- {0 ... 255}
--    IDELAY2_VALUE 		=> 0, 				-- {0 ... 255}
--    ODELAY_VALUE  		=> 0, 				-- {0 ... 255}
--    IDELAY_MODE   		=> "NORMAL", 			-- "NORMAL", "PCI"
--    SERDES_MODE   		=> "MASTER", 			-- <NONE>, MASTER, SLAVE
--    IDELAY_TYPE   		=> "FIXED", 	-- "DEFAULT", "DIFF_PHASE_DETECTOR", "FIXED", "VARIABLE_FROM_HALF_MAX", "VARIABLE_FROM_ZERO"
--    COUNTER_WRAPAROUND 	=> "STAY_AT_LIMIT", 		-- STAY_AT_LIMIT, WRAPAROUND
--    DELAY_SRC     		=> "IDATAIN" )			-- "IO", "IDATAIN", "ODATAIN"
--  port map (
--    IDATAIN  		=> ref_sdat_in, 		     -- input, data from master IOB
--    TOUT     		=> open, 			  -- output, tri-state signal to IOB
--    DOUT     		=> open, 			  -- output, data to IOB
--    T        		=> '1', 			     -- input, tri-state control from OLOGIC/OSERDES2, tie high fro input only mode
--    ODATAIN  		=> '0', 			     -- input, data from OLOGIC/OSERDES2
--    DATAOUT  		=> ref_sdat_dld_m, 	  -- output, delayed data 1 to ILOGIC/ISERDES2
--    DATAOUT2 		=> open, 			  -- output, data 2 to ILOGIC/ISERDES2, for PCI ?
--    IOCLK0   		=> '0', -- del_ioclk0, 			  -- input, High speed clock for calibration
--    
--    IOCLK1   		=> '0', 			     -- input, High speed clock for calibration
--    CLK      		=> '0', --clk,         -- input, Fabric clock (GCLK) for control signals
--    CAL      		=> '0', 	  -- input, Calibrate enable signal
--    INC      		=> '0', 			     -- input, increment / decrement delay counter
--    CE       		=> '0', 			     -- input, enable increment / decrement delay counter
--    RST      		=> '0',     -- input, Reset delay line to 1/2 max in this case
--    BUSY        => open) ;         -- output signal indicating sync circuit has finished / calibration has finished
--
--  ref_iodelay_s : IODELAY2 generic map(
--    DATA_RATE      		=> "SDR", 			-- <SDR>, DDR
--    SIM_TAPDELAY_VALUE	=> 37,  			-- nominal tap delay (sim parameter only)
--    IDELAY_VALUE  		=> 1, 				-- {0 ... 255}
--    IDELAY2_VALUE 		=> 0, 				-- {0 ... 255}
--    ODELAY_VALUE  		=> 0, 				-- {0 ... 255}
--    IDELAY_MODE   		=> "NORMAL", 			-- "NORMAL", "PCI"
--    SERDES_MODE   		=> "SLAVE", 			-- <NONE>, MASTER, SLAVE
--    IDELAY_TYPE   		=> "FIXED", 	-- "DEFAULT", "DIFF_PHASE_DETECTOR", "FIXED", "VARIABLE_FROM_HALF_MAX", "VARIABLE_FROM_ZERO"
--    COUNTER_WRAPAROUND 	=> "STAY_AT_LIMIT", 		-- STAY_AT_LIMIT, WRAPAROUND
--    DELAY_SRC     		=> "IDATAIN" )			-- "IO", "IDATAIN", "ODATAIN"
--  port map (
--    IDATAIN  		=> ref_sdat_in, 		     -- input, data from master IOB
--    TOUT     		=> open, 			  -- output, tri-state signal to IOB
--    DOUT     		=> open, 			  -- output, data to IOB
--    T        		=> '1', 			     -- input, tri-state control from OLOGIC/OSERDES2, tie high fro input only mode
--    ODATAIN  		=> '0', 			     -- input, data from OLOGIC/OSERDES2
--    DATAOUT  		=> ref_sdat_dld_s, 	  -- output, delayed data 1 to ILOGIC/ISERDES2
--    DATAOUT2 		=> open, 			  -- output, data 2 to ILOGIC/ISERDES2, for PCI ?
--    IOCLK0   		=> '0', --del_ioclk0, 			  -- input, High speed clock for calibration
--    IOCLK1   		=> '0', 			     -- input, High speed clock for calibration
--    CLK      		=> '0', --clk,         -- input, Fabric clock (GCLK) for control signals
--    CAL      		=> '0', 	  -- input, Calibrate enable signal
--    INC      		=> '0', 			     -- input, increment / decrement delay counter
--    CE       		=> '0', 			     -- input, enable increment / decrement delay counter
--    RST      		=> '0',       -- input, Reset delay line to 1/2 max in this case
--    BUSY      	=> open) ;         -- output signal indicating sync circuit has finished / calibration has finished

   
   ref_ISERDES2_master_inst : ISERDES2
     generic map (
        BITSLIP_ENABLE => TRUE,     -- Enable Bitslip Functionality (TRUE/FALSE)
        DATA_RATE      => "SDR",     -- Data-rate ("SDR" or "DDR")
        DATA_WIDTH     => S,         -- Parallel data width selection (2-8)
        INTERFACE_TYPE => "RETIMED", -- "NETWORKING", "NETWORKING_PIPELINED" or "RETIMED" 
                                     -- "RETIMED" to meet the timing requirements !!!
        SERDES_MODE => "MASTER"           -- "NONE", "MASTER" or "SLAVE" 
     )
     port map (
        CFB0      => open,      -- 1-bit output: Clock feed-through route output
        CFB1      => open,      -- 1-bit output: Clock feed-through route output
        DFB       => open,      -- 1-bit output: Feed-through clock output
        FABRICOUT => open, -- 1-bit output: Unsynchrnonized data output
        INCDEC    => open, --phase_incdec,      -- 1-bit output: Phase detector output
                      -- Q1 - Q4: 1-bit (each) output: Registered outputs to FPGA logic
        Q1       => ref_iserdes_pdat(4),
        Q2       => ref_iserdes_pdat(5),
        Q3       => ref_iserdes_pdat(6),
        Q4       => ref_iserdes_pdat(7),
        SHIFTOUT => ref_sdat_cascade, -- 1-bit output: Cascade output signal for master/slave I/O
        VALID    => open,--phase_valid,         -- 1-bit output: Output status of the phase detector
        BITSLIP  => '0', --bitslip,          -- 1-bit input: Bitslip enable input
        CE0      => '1',          -- 1-bit input: Clock enable input
        CLK0     => iserdes_ioclk,-- 1-bit input: I/O clock network input
        CLK1     => '0',          -- 1-bit input: Secondary I/O clock network input
        CLKDIV   => clk,          -- 1-bit input: FPGA logic domain clock input
        D        => ref_sdat_in, --ref_sdat_dld_m,   -- 1-bit input: Input data
        IOCE     => iserdes_stb,  -- 1-bit input: Data strobe input
        RST      => iserdes_rst,          -- 1-bit input: Asynchronous reset input
        SHIFTIN  => ref_pd_edge       -- 1-bit input: Cascade input signal for master/slave I/O
     );

    ref_sdat_ISERDES2_slave_inst : ISERDES2
     generic map (
        BITSLIP_ENABLE => TRUE,        -- Enable Bitslip Functionality (TRUE/FALSE)
        DATA_RATE => "SDR",             -- Data-rate ("SDR" or "DDR")
        DATA_WIDTH => S,                -- Parallel data width selection (2-8)
        INTERFACE_TYPE => "RETIMED",    -- "NETWORKING", "NETWORKING_PIPELINED" or "RETIMED"
                                        -- "RETIMED" to meet the timing requirements !!!
        SERDES_MODE => "SLAVE"          -- "NONE", "MASTER" or "SLAVE" 
     )
     port map (
        CFB0     => open,    -- 1-bit output: Clock feed-through route output
        CFB1     => open,    -- 1-bit output: Clock feed-through route output
        DFB      => open,    -- 1-bit output: Feed-through clock output
        FABRICOUT=> open,    -- 1-bit output: Unsynchrnonized data output
        INCDEC   => open,    -- 1-bit output: Phase detector output
                             -- Q1 - Q4: 1-bit (each) output: Registered outputs to FPGA logic
        Q1       => ref_iserdes_pdat(0),
        Q2       => ref_iserdes_pdat(1),
        Q3       => ref_iserdes_pdat(2),
        Q4       => ref_iserdes_pdat(3),
        SHIFTOUT => ref_pd_edge,    -- 1-bit output: Cascade output signal for master/slave I/O
        VALID    => open,   -- 1-bit output: Output status of the phase detector
        BITSLIP  => '0',     -- 1-bit input: Bitslip enable input
        CE0      => '1',       -- 1-bit input: Clock enable input
        CLK0     => iserdes_ioclk, -- 1-bit input: I/O clock network input
        CLK1     => '0',        -- 1-bit input: Secondary I/O clock network input
        CLKDIV   => clk,     -- 1-bit input: FPGA logic domain clock input
        D        => ref_sdat_in, --ref_sdat_dld_s,    -- 1-bit input: Input data
        IOCE     => iserdes_stb, -- 1-bit input: Data strobe input
        RST      => iserdes_rst,   -- 1-bit input: Asynchronous reset input
        SHIFTIN  => ref_sdat_cascade   -- 1-bit input: Cascade input signal for master/slave I/O
     );

--- iodelay and iserdes instantiation for the signal to be calibrated ----------------------------------------------------

--  cal_iodelay_m : IODELAY2 generic map(
--    DATA_RATE      		=> "SDR", 			-- <SDR>, DDR
--    SIM_TAPDELAY_VALUE	=> 37,  			-- nominal tap delay (sim parameter only)
--    IDELAY_VALUE  		=> 0, 				-- {0 ... 255}
--    IDELAY2_VALUE 		=> 0, 				-- {0 ... 255}
--    ODELAY_VALUE  		=> 0, 				-- {0 ... 255}
--    IDELAY_MODE   		=> "NORMAL", 			-- "NORMAL", "PCI"
--    SERDES_MODE   		=> "MASTER", 			-- <NONE>, MASTER, SLAVE
--    IDELAY_TYPE   		=> "FIXED", 	-- "DEFAULT", "DIFF_PHASE_DETECTOR", "FIXED", "VARIABLE_FROM_HALF_MAX", "VARIABLE_FROM_ZERO"
--    COUNTER_WRAPAROUND 	=> "STAY_AT_LIMIT", 		-- STAY_AT_LIMIT, WRAPAROUND
--    DELAY_SRC     		=> "IDATAIN" )			-- "IO", "IDATAIN", "ODATAIN"
--  port map (
--    IDATAIN  		=> cal_sdat_in, 		     -- input, data from master IOB
--    TOUT     		=> open, 			  -- output, tri-state signal to IOB
--    DOUT     		=> open, 			  -- output, data to IOB
--    T        		=> '1', 			     -- input, tri-state control from OLOGIC/OSERDES2, tie high fro input only mode
--    ODATAIN  		=> '0', 			     -- input, data from OLOGIC/OSERDES2
--    DATAOUT  		=> cal_sdat_dld_m, 	  -- output, delayed data 1 to ILOGIC/ISERDES2
--    DATAOUT2 		=> open, 			  -- output, data 2 to ILOGIC/ISERDES2, for PCI ?
--    IOCLK0   		=> del_ioclk0, 			  -- input, High speed clock for calibration
--    
--    IOCLK1   		=> '0', 			     -- input, High speed clock for calibration
--    CLK      		=> clk,         -- input, Fabric clock (GCLK) for control signals
--    CAL      		=> '0', 	  -- input, Calibrate enable signal
--    INC      		=> '0', 			     -- input, increment / decrement delay counter
--    CE       		=> '0', 			     -- input, enable increment / decrement delay counter
--    RST      		=> '0',     -- input, Reset delay line to 1/2 max in this case
--    BUSY        => open) ;         -- output signal indicating sync circuit has finished / calibration has finished
--
--  cal_iodelay_s : IODELAY2 generic map(
--    DATA_RATE      		=> "SDR", 			-- <SDR>, DDR
--    SIM_TAPDELAY_VALUE	=> 37,  			-- nominal tap delay (sim parameter only)
--    IDELAY_VALUE  		=> 0, 				-- {0 ... 255}
--    IDELAY2_VALUE 		=> 0, 				-- {0 ... 255}
--    ODELAY_VALUE  		=> 0, 				-- {0 ... 255}
--    IDELAY_MODE   		=> "NORMAL", 			-- "NORMAL", "PCI"
--    SERDES_MODE   		=> "SLAVE", 			-- <NONE>, MASTER, SLAVE
--    IDELAY_TYPE   		=> "FIXED", 	-- "DEFAULT", "DIFF_PHASE_DETECTOR", "FIXED", "VARIABLE_FROM_HALF_MAX", "VARIABLE_FROM_ZERO"
--    COUNTER_WRAPAROUND 	=> "STAY_AT_LIMIT", 		-- STAY_AT_LIMIT, WRAPAROUND
--    DELAY_SRC     		=> "IDATAIN" )			-- "IO", "IDATAIN", "ODATAIN"
--  port map (
--    IDATAIN  		=> cal_sdat_in, 		     -- input, data from master IOB
--    TOUT     		=> open, 			  -- output, tri-state signal to IOB
--    DOUT     		=> open, 			  -- output, data to IOB
--    T        		=> '1', 			     -- input, tri-state control from OLOGIC/OSERDES2, tie high fro input only mode
--    ODATAIN  		=> '0', 			     -- input, data from OLOGIC/OSERDES2
--    DATAOUT  		=> cal_sdat_dld_s, 	  -- output, delayed data 1 to ILOGIC/ISERDES2
--    DATAOUT2 		=> open, 			  -- output, data 2 to ILOGIC/ISERDES2, for PCI ?
--    IOCLK0   		=> del_ioclk0, 			  -- input, High speed clock for calibration
--    IOCLK1   		=> '0', 			     -- input, High speed clock for calibration
--    CLK      		=> clk,         -- input, Fabric clock (GCLK) for control signals
--    CAL      		=> '0', 	  -- input, Calibrate enable signal
--    INC      		=> '0', 			     -- input, increment / decrement delay counter
--    CE       		=> '0', 			     -- input, enable increment / decrement delay counter
--    RST      		=> '0',       -- input, Reset delay line to 1/2 max in this case
--    BUSY      	=> open) ;         -- output signal indicating sync circuit has finished / calibration has finished

   
   cal_ISERDES2_master_inst : ISERDES2
     generic map (
        BITSLIP_ENABLE => TRUE,     -- Enable Bitslip Functionality (TRUE/FALSE)
        DATA_RATE      => "SDR",     -- Data-rate ("SDR" or "DDR")
        DATA_WIDTH     => S,         -- Parallel data width selection (2-8)
        INTERFACE_TYPE => "RETIMED", -- "NETWORKING", "NETWORKING_PIPELINED" or "RETIMED" 
                                     -- "RETIMED" to meet the timing requirements !!!
        SERDES_MODE => "MASTER"           -- "NONE", "MASTER" or "SLAVE" 
     )
     port map (
        CFB0      => open,      -- 1-bit output: Clock feed-through route output
        CFB1      => open,      -- 1-bit output: Clock feed-through route output
        DFB       => open,      -- 1-bit output: Feed-through clock output
        FABRICOUT => open, -- 1-bit output: Unsynchrnonized data output
        INCDEC    => open, --phase_incdec,      -- 1-bit output: Phase detector output
                      -- Q1 - Q4: 1-bit (each) output: Registered outputs to FPGA logic
        Q1       => cal_iserdes_pdat(4),
        Q2       => cal_iserdes_pdat(5),
        Q3       => cal_iserdes_pdat(6),
        Q4       => cal_iserdes_pdat(7),
        SHIFTOUT => cal_sdat_cascade, -- 1-bit output: Cascade output signal for master/slave I/O
        VALID    => open,--phase_valid,         -- 1-bit output: Output status of the phase detector
        BITSLIP  => '0', --bitslip,          -- 1-bit input: Bitslip enable input
        CE0      => '1',          -- 1-bit input: Clock enable input
        CLK0     => iserdes_ioclk,-- 1-bit input: I/O clock network input
        CLK1     => '0',          -- 1-bit input: Secondary I/O clock network input
        CLKDIV   => clk,          -- 1-bit input: FPGA logic domain clock input
        D        => cal_sdat_in,  --cal_sdat_dld_m,   -- 1-bit input: Input data
        IOCE     => iserdes_stb,  -- 1-bit input: Data strobe input
        RST      => iserdes_rst,          -- 1-bit input: Asynchronous reset input
        SHIFTIN  => cal_pd_edge       -- 1-bit input: Cascade input signal for master/slave I/O
     );

    cal_sdat_ISERDES2_slave_inst : ISERDES2
     generic map (
        BITSLIP_ENABLE => TRUE,        -- Enable Bitslip Functionality (TRUE/FALSE)
        DATA_RATE => "SDR",             -- Data-rate ("SDR" or "DDR")
        DATA_WIDTH => S,                -- Parallel data width selection (2-8)
        INTERFACE_TYPE => "RETIMED",    -- "NETWORKING", "NETWORKING_PIPELINED" or "RETIMED"
                                        -- "RETIMED" to meet the timing requirements !!!
        SERDES_MODE => "SLAVE"          -- "NONE", "MASTER" or "SLAVE" 
     )
     port map (
        CFB0     => open,    -- 1-bit output: Clock feed-through route output
        CFB1     => open,    -- 1-bit output: Clock feed-through route output
        DFB      => open,    -- 1-bit output: Feed-through clock output
        FABRICOUT=> open,    -- 1-bit output: Unsynchrnonized data output
        INCDEC   => open,    -- 1-bit output: Phase detector output
                             -- Q1 - Q4: 1-bit (each) output: Registered outputs to FPGA logic
        Q1       => cal_iserdes_pdat(0),
        Q2       => cal_iserdes_pdat(1),
        Q3       => cal_iserdes_pdat(2),
        Q4       => cal_iserdes_pdat(3),
        SHIFTOUT => cal_pd_edge,    -- 1-bit output: Cascade output signal for master/slave I/O
        VALID    => open,   -- 1-bit output: Output status of the phase detector
        BITSLIP  => '0',     -- 1-bit input: Bitslip enable input
        CE0      => '1',       -- 1-bit input: Clock enable input
        CLK0     => iserdes_ioclk, -- 1-bit input: I/O clock network input
        CLK1     => '0',        -- 1-bit input: Secondary I/O clock network input
        CLKDIV   => clk,     -- 1-bit input: FPGA logic domain clock input
        D        => cal_sdat_in,  -- cal_sdat_dld_s,    -- 1-bit input: Input data
        IOCE     => iserdes_stb, -- 1-bit input: Data strobe input
        RST      => iserdes_rst,   -- 1-bit input: Asynchronous reset input
        SHIFTIN  => cal_sdat_cascade   -- 1-bit input: Cascade input signal for master/slave I/O
     );


-- do pulse shaping by leading edge ----------------------------------------------- 


 REF_PULSE_SHAPE: process (clk)
    begin
      if rising_edge(clk) then   -- start pulse shaping, find the LH-edge and a generate a single bit word
       if pulse_ena = '1' then      
        ref_iserdes_pdat_p7 <= ref_iserdes_pdat(7);
        if ref_iserdes_pdat_p7='0' and ref_iserdes_pdat(0)='1' then ref_pulse_pdat <= X"80"; end if;
        if ref_iserdes_pdat(0)='0' and ref_iserdes_pdat(1)='1' then ref_pulse_pdat <= X"40"; end if;
        if ref_iserdes_pdat(1)='0' and ref_iserdes_pdat(2)='1' then ref_pulse_pdat <= X"20"; end if;
        if ref_iserdes_pdat(2)='0' and ref_iserdes_pdat(3)='1' then ref_pulse_pdat <= X"10"; end if;
        if ref_iserdes_pdat(3)='0' and ref_iserdes_pdat(4)='1' then ref_pulse_pdat <= X"08"; end if;
        if ref_iserdes_pdat(4)='0' and ref_iserdes_pdat(5)='1' then ref_pulse_pdat <= X"04"; end if;
        if ref_iserdes_pdat(5)='0' and ref_iserdes_pdat(6)='1' then ref_pulse_pdat <= X"02"; end if;
        if ref_iserdes_pdat(6)='0' and ref_iserdes_pdat(7)='1' then ref_pulse_pdat <= X"01"; end if;
       else
        ref_iserdes_pdat_p7    <= '0';
        ref_pulse_pdat         <= X"00";
        ref_pulse_pdat_prev    <= (others =>'0');
      end if;  
      ref_pulse_pdat_prev    <= ref_pulse_pdat(4 to 7);
     end if; -- if rising_edge(clk) 
    end process REF_PULSE_SHAPE;

 CAL_PULSE_SHAPE: process (clk)
    begin
      if rising_edge(clk) then   -- start pulse shaping, find the LH-edge and a generate a single bit word
       if pulse_ena = '1' then      
        cal_iserdes_pdat_p7 <= cal_iserdes_pdat(7);
        if cal_iserdes_pdat_p7='0' and cal_iserdes_pdat(0)='1' then cal_pulse_pdat <= X"80"; end if;
        if cal_iserdes_pdat(0)='0' and cal_iserdes_pdat(1)='1' then cal_pulse_pdat <= X"40"; end if;
        if cal_iserdes_pdat(1)='0' and cal_iserdes_pdat(2)='1' then cal_pulse_pdat <= X"20"; end if;
        if cal_iserdes_pdat(2)='0' and cal_iserdes_pdat(3)='1' then cal_pulse_pdat <= X"10"; end if;
        if cal_iserdes_pdat(3)='0' and cal_iserdes_pdat(4)='1' then cal_pulse_pdat <= X"08"; end if;
        if cal_iserdes_pdat(4)='0' and cal_iserdes_pdat(5)='1' then cal_pulse_pdat <= X"04"; end if;
        if cal_iserdes_pdat(5)='0' and cal_iserdes_pdat(6)='1' then cal_pulse_pdat <= X"02"; end if;
        if cal_iserdes_pdat(6)='0' and cal_iserdes_pdat(7)='1' then cal_pulse_pdat <= X"01"; end if;
       else
        cal_iserdes_pdat_p7    <= '0';
        cal_pulse_pdat         <= X"00";
        cal_pulse_pdat_prev    <= (others =>'0');
      end if;  
      cal_pulse_pdat_prev    <= cal_pulse_pdat(4 to 7);
     end if; -- if rising_edge(clk) 
    end process CAL_PULSE_SHAPE;
    
-- generate signals to control the io_delay
    
   INCDEC_GEN_stm: process (reset, clk) 
     begin
     if clk'event and clk = '1' then
      if reset = '1' then
        inc_delay     <= '0';
        dec_delay     <= '0';
        phase_aligned <= '0';
        pulse_ena     <= '0';
        pulse_state   <= PULSE_IDLE;
      else 
      
   PULSE_SM : case (pulse_state) is
   
      when PULSE_IDLE => 
        inc_delay     <= '0';
        dec_delay     <= '0';
        phase_aligned <= '0';
        pulse_ena     <= '1';
        if cal_pulse_pdat /= X"00" or ref_pulse_pdat /= X"00" then
          ref_pulse_pdat_incl_prev  <= (ref_pulse_pdat_prev & ref_pulse_pdat);
          cal_pulse_pdat_incl_prev  <= (cal_pulse_pdat_prev & cal_pulse_pdat);
          pulse_ena                 <= '0';
          pulse_state               <= INCDEC_GEN;
        end if;
       
      when INCDEC_GEN =>
       if ref_pulse_pdat_incl_prev /= X"000" and cal_pulse_pdat_incl_prev /= X"000" then
        if ref_pulse_pdat_incl_prev > cal_pulse_pdat_incl_prev then 
         dec_delay     <= '1';
        end if; 
        if ref_pulse_pdat_incl_prev < cal_pulse_pdat_incl_prev then 
         inc_delay     <= '1';
        end if;
        if ref_pulse_pdat_incl_prev = cal_pulse_pdat_incl_prev then
         phase_aligned <= '1';
        end if;
       end if;
       pulse_state <= PULSE_IDLE;
       
       end case PULSE_SM;
       
      end if;-- reset = '1' 
     end if; -- clk'event and clk = '1' 
    end process  INCDEC_GEN_stm;    
               

-- controlling the input delay of the signal to get calibated ---------------------------------- 

 CAL_ISERDES: process (clk)
    begin
     if clk'event and clk = '1' then
       if reset = '1' then
         state        <= CAL_IDLE;
         iodelay_cal  <= '0';
         iodelay_inc  <= '0';
         iodelay_ce   <= '0';
         cal_busy     <= '0';
         cal_err      <= '0';
         cal_done     <= '0';
         del_count    <= (others => '0');
         in_phase_ct  <= (others => '0');
         inc_plus     <= '0';
       else
         
   CAL_SM : case (state) is
   
          when CAL_IDLE => 
            iodelay_cal  <= '0';
            iodelay_inc  <= '0';
            iodelay_ce   <= '0';
            cal_busy     <= '0';
            cal_done     <= '0';
            del_count    <= X"3f";
            in_phase_ct  <= (others => '0');
            inc_plus     <= '0';
            if cal_start = '1' then  -- otherwise save the last calibration status
             cal_busy    <= '1';        -- let the software wait until calibration or delay load              
             cal_err     <= '0';
             iodelay_cal <= '1';
             state       <= DEL_CAL_WAIT;
            end if;
            
         when DEL_CAL_WAIT =>          -- wait for the command being accepted
            if iodelay_busy = '1' then 
              iodelay_cal  <= '0';
              state        <= DEL_CAL;
            end if;  
              
          when DEL_CAL =>             -- wait for calbration-readiness
            if iodelay_busy = '0' then -- both busy are gone
              iodelay_rst <= '1';             
              state       <= DEL_RST;
            end if;    

          when DEL_RST =>             -- reset the delay channel
            iodelay_rst <= '0';
            if iodelay_busy = '0' then
              state      <= CAL_BUSY_WAIT;
            end if;        
             
          when CAL_PHASE_CHECK => 
            if del_count = X"00" or del_count = X"ff" then
             cal_err <= '1';
             state <= CAL_IDLE;
            else   
              if dec_delay = '1' then
                iodelay_inc  <= '0';
                iodelay_ce   <= '1';              -- decrement the delay
                if  del_count /= X"00" then
                 del_count <= del_count - 1;
                end if;                 
                if in_phase_ct /= 0 then  
                 in_phase_ct <= in_phase_ct - 1; 
                end if;
                state        <= CAL_BUSY_ON;
              end if; -- dec_delay = '1'
              if inc_delay = '1' then
                iodelay_inc  <= '1';
                iodelay_ce   <= '1';              -- increment the delay
                if  del_count /= X"ff" then
                 del_count <= del_count + 1;
                end if; 
                if in_phase_ct /= 0 then  
                 in_phase_ct <= in_phase_ct - 1; 
                end if;
                state        <= CAL_BUSY_ON;
              end if; -- inc_delay = '1' 
              if phase_aligned = '1' then              
               if in_phase_ct /= (ALIGNMENT_CNT-EXTRA_INC_CNT) then        
                in_phase_ct <= in_phase_ct + 1;   
               else
                inc_plus <= '1';              -- set flag 
                state    <= CAL_INC_PLUS;     -- delay adjusted now
               end if; 
              end if;-- phase_aligned = '1'
             end if; --  del_count = X"00" or del_count "X"ff" 

          when CAL_INC_PLUS  =>                   -- do 6 more increments for perfect alignment
               if in_phase_ct /= ALIGNMENT_CNT then        
                in_phase_ct  <= in_phase_ct + 1;
                iodelay_inc  <= '1';
                iodelay_ce   <= '1';              -- increment the delay
                if  del_count /= X"ff" then
                 del_count <= del_count + 1;
                end if;
                state      <= CAL_BUSY_ON;                
               else
                state    <= CAL_IDLE;               -- delay adjusted now
                cal_done <= '1';
               end if; --in_phase_ct /= ALIGNMENT_CNT

          when CAL_BUSY_ON =>
             iodelay_ce    <= '0';
             --if ((iodelay_busy_m and iodelay_busy_s) = '1') then
               state        <= CAL_BUSY_WAIT;
             --end if;
             
          when CAL_BUSY_WAIT =>
             if iodelay_busy = '0' then
               busy_wait_ct <= B"100"; -- required for clean ce to busy handling
               state        <= CAL_DEL_ADJ_WAIT;
             end if;

          when CAL_DEL_ADJ_WAIT =>
             if busy_wait_ct = B"000" then
              if inc_plus = '0' then
               state <= CAL_PHASE_CHECK;
              else
               state <= CAL_INC_PLUS;
              end if; -- inc_plus = '0'              
             else
               busy_wait_ct  <= busy_wait_ct - 1;
             end if; 
             
        end case CAL_SM;
      end if;
    end if;
   end process CAL_ISERDES;
   
        cal_iodelay_rst  <= iodelay_rst ;   -- iodelay_ce, to control the delay of the signal to be calibrated 
        cal_iodelay_cal  <= iodelay_cal ;   -- iodelay_ce, to control the delay of the signal to be calibrated
        cal_iodelay_ce   <= iodelay_ce  ;   -- iodelay_ce, to control the delay of the signal to be calibrated
        cal_iodelay_inc  <= iodelay_inc ;   -- iodelay_inc, to control the delay of the signal to be calibrated
        iodelay_busy     <= cal_iodelay_busy ;    -- iodelay_busy, from the signal to be calibrated
 
 
end arch_iserdes_ph_ctrl_02;
