-------------------------------------------------------
-- Design Name : spi_slave_8bit_02
-- File Name   : spi_slave_8bit_02.vhd
-- Device      : Spartan 6
-- Function    : CTDB SPI bus slave
-- Coder       : K.-H. Sulanke, DESY, 2017-02-09
-------------------------------------------------------
-- 2016-09-07  high speed system clock for synchronization implemented
--             using delayed spi_sclk lh-edge for data sampling
-- 2017-02-08  SPI read bug fixed and SPI_MISO tristate replaced by '0'
--             note, when master write cycle, slave rx data are sampled at spi_clk_lh (!!!)
--             instead of ..._hl 

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all ;

library unisim ;
use unisim.vcomponents.all ;

entity spi_slave_8bit_02 is
  port
   (
     clk             : in  std_logic ;  -- high speed system clock
     spi_reset       : in  std_logic ;  -- synchronous spi_reset
     spi_txd         : in  std_logic_vector (7 downto 0); -- data to be sent
     --spi_tx_done     : out std_logic;   -- tx data sent, single pulse 
     spi_rxd         : out std_logic_vector (7 downto 0); -- data received
     spi_rxd_val     : out std_logic;   -- Rx data vaild, single pulse
     spi_addr        : out std_logic_vector ( 6 downto 0); -- register address
     spi_addr_val    : out std_logic;   -- Rx data vaild, single pulse
     spi_wr_nrd      : out  std_logic ; -- spi cycle is active
     spi_sync        : in  std_logic ;  -- backplane bus, low active sync signal
     spi_sclk        : in  std_logic ;  -- backplane bus, SPI clock
     spi_mosi        : in  std_logic ;  -- backplane bus, serial SPI tx-data
     spi_miso        : out std_logic    -- backplane bus, serial SPI rx-data
     --spi_miso_trst   : out std_logic    -- to disable the LVDS bus driver
   );
end entity;

architecture spi_slave_8bit_02_arch of spi_slave_8bit_02 is

 
  signal spi_sclk_lh    : std_logic; -- rising edge
  --signal spi_sclk_hl    : std_logic; -- falling edge
  signal spi_sclk_pipe  : std_logic_vector (2 downto 0) := B"000"; -- 
  signal reg_wr         : std_logic;
  signal reg_rd         : std_logic;
  signal tx_rg          : std_logic_vector (7 downto 0) := X"00"; --
  signal rx_rg          : std_logic_vector (7 downto 0) := X"00"; --
  signal spi_addr_nd    : std_logic_vector (6 downto 0) := B"0000000"; --
  signal spi_rxd_rg     : std_logic_vector (7 downto 0) := X"00"; --
  signal sl_cnt         : integer range 0 to 16 := 0;
  signal spi_miso_trst  : std_logic;
 
--  type state_type is (IDLE, GET_ADDR, REC_DATA, SEND_DATA, SPI_DONE);
--  signal state : state_type := IDLE;

begin

  GET_SPI_CLK_EDGES: process (spi_reset, clk)
  begin
   if rising_edge(clk) then
    if spi_reset = '1' then
     spi_sclk_lh <= '0';
     --spi_sclk_hl <= '0';
     spi_sclk_pipe <= B"000";
    else
     spi_sclk_pipe(0) <= spi_sclk;    
     spi_sclk_pipe(1) <= spi_sclk_pipe(0);    
     spi_sclk_pipe(2) <= spi_sclk_pipe(1); 
     if spi_sclk_pipe(1) = '1' and spi_sclk_pipe(2) = '0' then
       spi_sclk_lh <= '1';
     else
       spi_sclk_lh <= '0';
     end if;
--     if spi_sclk_pipe(1) = '0' and spi_sclk_pipe(2) = '1' then
--       spi_sclk_hl <= '1';
--     else
--       spi_sclk_hl <= '0';
--     end if;
    end if;-- spi_reset = '1'
   end if; --rising_edge(clk)
  end process GET_SPI_CLK_EDGES;  

  SPI_REC: process (clk, spi_sync) -- use LH-edge for data sampling
   begin
    if rising_edge(clk) then
       --spi_tx_done    <= '0';  -- to get a single pulse
       spi_rxd_val    <= '0';  -- to get a single pulse
       spi_addr_val   <= '0';  -- to get a single pulse
     if (spi_sync = '1') then
       sl_cnt         <= 0;
       reg_wr         <= '0';
       reg_rd         <= '0';
       spi_miso_trst  <= '1';
     else
      if spi_sclk_lh = '1' then
        sl_cnt             <= sl_cnt + 1;
        rx_rg(0)           <= spi_mosi;
        rx_rg(7 downto 1)  <= rx_rg(6 downto 0);
        if (sl_cnt = 1) then
         reg_wr      <= rx_rg(0);
        end if;       
        if (sl_cnt = 7) then 
          spi_addr_val <= '1';
          spi_addr_nd  <= rx_rg(5 downto 0) & spi_mosi;
        end if;
        if (sl_cnt = 7) then       
          if (reg_wr = '0') then
           reg_rd     <= '1';
          end if;
        end if; --(sl_cnt = 7)
--        if (sl_cnt = 15) then
--         if (reg_wr = '1') and (spi_sclk_hl='1') then
--           spi_rxd_val   <= '1';
--         else
--           spi_rxd_val   <= '0';
--         end if;
--         if (reg_wr = '0') and (spi_sclk_hl='1') then 
--           spi_tx_done   <= '1'; 
--         else
--           spi_tx_done   <= '0';         
--         end if;
--        end if; -- (sl_cnt = 15)
        if (sl_cnt = 15) then
         if (reg_wr = '1') then
           spi_rxd_val   <= '1';
         --else
           --spi_tx_done   <= '1';       
         end if;
        end if; -- (sl_cnt = 15) 
       end if; -- spi_sclk_lh = '1'        
       if (reg_rd = '1') then
        if sl_cnt > 8 then spi_miso_trst <= '0'; end if; -- sl_cnt > 8
         if spi_sclk_lh = '1' then
          if (sl_cnt = 8) then
            tx_rg   <= spi_txd;
          else
            tx_rg(7 downto 1) <= tx_rg(6 downto 0);       
          end if;-- (sl_cnt = 8)  
         end if; -- spi_sclk_hl = '1'
        end if; -- (reg_rd = '1')
    end if; -- (spi_sync = '1')      
   end if; -- (rising_edge(clk)) 
  end process SPI_REC;

   spi_rxd       <= rx_rg;
   spi_addr      <= spi_addr_nd;   
--   spi_miso      <= tx_rg(7) when spi_miso_trst ='0' else 'Z';
   spi_miso      <= tx_rg(7) when spi_miso_trst ='0' else '0';
   spi_wr_nrd    <= reg_wr;
  
 end architecture spi_slave_8bit_02_arch;
