-------------------------------------------------------
-- Design Name : sw_led_ctrl_01
-- File Name   : sw_led_ctrl_01.vhd
-- Device      : Spartan 6
-- Function    : SW1/2 debouncing and LED (V3) control 
-- Coder       : K.-H. Sulanke, DESY, 2017-03-28
-------------------------------------------------------
-- 2017-03-28

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all ;

library unisim ;
use unisim.vcomponents.all ;

entity sw_led_ctrl_01 is  
  port (
        reset 		     : in std_logic ;   -- reset (active high)
        clk            : in std_logic ;   -- global clock input
        nSW1           : in  std_logic; -- on board push button, can be used for calibration
       -- nSW2           : in  std_logic; -- on board push button, pulls signal to GND
        nLED_GREEN     : out std_logic; -- '0' = LED V3-green is on
        nLED_RED       : out std_logic; -- '0' = LED V3-red is on
        cal_start      :  out std_logic ;   -- start calibration cycle, single pulse
        cal_done       :  in  std_logic;   -- calibration status
        cal_err        :  in  std_logic    -- calibration error, gets cleared by next cal_start
        );
  end sw_led_ctrl_01 ;

architecture arch_sw_led_ctrl_01 of sw_led_ctrl_01 is

-- 
	signal sw1         : std_logic_vector(3 downto 0) := X"0"; 
  signal sw1_pressed : boolean := false;
  signal led_green   : std_logic := '0';
  signal led_red     : std_logic := '0';
  signal ms_divider  : integer range 0 to 125000 := 0;
  --signal ms_pulse    : std_logic;
  signal ms_ct_ena   : std_logic;
  signal ms_ct       : std_logic_vector(9 downto 0) := (others => '0');   
   
  type state_type is (CAL_IDLE, CAL_RUN, CAL_RDY, CAL_ERROR, CAL_WAIT);
   signal state : state_type := CAL_IDLE; 
 
  
begin

 MSEC_PULSE : process (clk, reset)
  begin
   if rising_edge(clk) then 
    if reset = '1' then
     ms_divider <= 0;
     ms_ct      <= (others => '0');
     --ms_pulse   <= '0';
    else
     if ms_ct_ena = '1' then
      if ms_divider = 124999 then  -- get 1ms pulse from 125 MHZ clock
       ms_divider  <= 0;
       --ms_pulse    <= '1';
       if ms_ct /= B"11_1110_0111" then -- 999 ms
        ms_ct  <= ms_ct + 1;
       else
        ms_ct  <= (others => '0');
       end if;          
      else      
       ms_divider <= ms_divider + 1;
       --ms_pulse   <= '0';   -- for single pulses
      end if; -- ms_divider = 124999
     else
      --ms_pulse    <= '0';
      ms_divider  <= 0;
      ms_ct       <= (others => '0');      
     end if; -- ms_cnt_ena = '1'
    end if; -- reset = '1'
   end if; -- rising_edge(clk)
  end process MSEC_PULSE;   
   

 SWx_DEBOUNCE: process (clk, reset)
    begin
      if rising_edge(clk) then   -- debounce SW1 and SW2
       if reset = '1' then
        sw1 <= B"0000";
       else
        sw1(0)          <= not nSW1;
        sw1(3 downto 1) <= sw1(2 downto 0);  
        if sw1(3 downto 2) = B"10" then -- switch is being released
         sw1_pressed <= true;
        else
         sw1_pressed <= false;        
        end if; -- sw1(3 downto 2) = B"01"
       end if; -- reset = '1' 
      end if; --rising_edge(clk)     
    end process SWx_DEBOUNCE;
    
   CAL_CTRL_stm: process (reset, clk) -- 
     begin
     if rising_edge(clk) then
      if reset = '1' then
        led_green     <= '0';
        led_red       <= '0';
        ms_ct_ena     <= '0';
        cal_start     <= '0';
        state         <= CAL_IDLE;
      else 
      
    SWx_SM : case (state) is
   
      when CAL_IDLE => 
        led_green     <= '0';
        led_red       <= '0';
        ms_ct_ena     <= '0';
        cal_start     <= '0';
        if sw1_pressed then
         cal_start    <= '1';      -- start calibration cycle now
         state        <= CAL_RUN;
        end if;
       
      when CAL_RUN =>
        cal_start    <= '0';           -- get single pulse
        ms_ct_ena    <= '1';           -- start milli-seconds counter 
        led_green    <= not ms_ct(8);  -- flashing green
        if cal_err = '1' then
         ms_ct_ena <= '0';  -- clear the counter
         state     <= CAL_ERROR;
        elsif cal_done = '1' then
         state     <= CAL_RDY;
        end if;

     when CAL_RDY =>          
        ms_ct_ena  <= '0';    -- clear the counter
        led_green  <= '1';    -- green permanent on
        if sw1_pressed then   -- acknowledge calibration readiness
         led_green <= '0';
         state     <= CAL_WAIT; 
        end if;

      when CAL_ERROR =>
        ms_ct_ena     <= '1';
        led_green     <= '0';
        led_red       <= not ms_ct(9); -- start with on, flashing red with 1Hz
        if sw1_pressed then            -- acknowledge calibration failure 
         led_red  <= '0';
         state    <= CAL_WAIT; 
        end if;

      when CAL_WAIT =>          -- wait half second before accepting next SW1 pressing
        ms_ct_ena  <= '1'; 
        if ms_ct(9) = '1' then
         state     <= CAL_IDLE; 
        end if;
       
      end case SWx_SM;
       
      end if; -- reset = '1' 
     end if; -- rising_edge(clk) 
    end process  CAL_CTRL_stm;    
 
  nLED_GREEN <= '0' when led_green = '1' else 'Z';
  nLED_RED   <= '0' when led_red   = '1' else 'Z';


end arch_sw_led_ctrl_01;